"""SANS submodule.

Module for calculation of Small Angle Neutron Scattering related
quantities.
"""

import discretisedfield as df
import numpy as np
from scipy.spatial.transform import Rotation


def cross_section(field, /, method, polarisation=(0, 0, 1)):
    r""" Calculation of scattering cross sections.

    The scattering cross sections can be calculated using

    .. math::
        \frac{d\sum}{d\Omega} \propto |{\bf Q} \cdot {\bf \sigma}|^2,

    where :math:`{\bf \sigma}` is the Pauli vector

    .. math::
        {\bf \sigma} = \begin{bmatrix} \sigma_x \\
                                       \sigma_y \\
                                       \sigma_z \end{bmatrix}

    and

    .. math::
        \begin{align}
            \sigma_x &= \begin{pmatrix} 0 & 1 \\ 1 & 0 \end{pmatrix}, \\
            \sigma_y &= \begin{pmatrix} 0 & -i \\ i & 0 \end{pmatrix}, \\
            \sigma_z &= \begin{pmatrix} 1 & 0 \\ 0 & -1 \end{pmatrix}.
        \end{align}

    :math:`{\bf Q}` is the magnetic interaction vector given by

    .. math::
        \begin{equation}
            {\bf Q} = \hat{\bf q} \times \widetilde{\bf M} \times \hat{\bf q}
        \end{equation}

    :math:`\hat{\bf q}` is the unit scattering vector and
    :math:`\widetilde{\bf M}` is the Fourier transform of the magnetisation.
    The magnetic interaction vector is dependent on the scattering geometry
    and the scattering vector is defined as

    .. math::

        \begin{equation}
            {\bf q} = {\bf k}_1 - {\bf k}_0.
        \end{equation}

    The spin-flip and non-spin-flip neutron scattering cross sections can be
    calculated for specific scattering geometries using

    .. math::
        \begin{equation}
         \frac{d\sum}{d\Omega} = \begin{pmatrix}
                                    \frac{d\sum^{++}}{d\Omega} &
                                    \frac{d\sum^{-+}}{d\Omega}\\
                                    \frac{d\sum^{+-}}{d\Omega} &
                                    \frac{d\sum^{--}}{d\Omega}
                                 \end{pmatrix}.
        \end{equation}

    The spin based cross sections then be combined in order to get the half
    polarised cross sections

    .. math::

        \begin{align}
            \frac{d\sum^{+}}{d\Omega} &= \frac{d\sum^{++}}{d\Omega} +
            \frac{d\sum^{+-}}{d\Omega}, \\
            \frac{d\sum^{-}}{d\Omega} &= \frac{d\sum^{--}}{d\Omega} +
            \frac{d\sum^{-+}}{d\Omega}.
        \end{align}

    These can further be combined to get the unpolarised cross section

    .. math::
        \begin{align}
            \frac{d\sum}{d\Omega} &= \frac{1}{2} \left(
            \frac{d\sum^{+}}{d\Omega} + \frac{d\sum^{-}}{d\Omega} \right), \\
            \frac{d\sum}{d\Omega} &= \frac{1}{2}
            \left( \frac{d\sum^{++}}{d\Omega} + \frac{d\sum^{+-}}{d\Omega} +
            \frac{d\sum^{--}}{d\Omega} + \frac{d\sum^{-+}}{d\Omega} \right).
        \end{align}

    Note: Similar to the magnetisation, the value of the cross section 
    exists within the whole cell and therefore will change based on the size
    of the cell and region in the input field.  

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.
    method : str
        Used to select the relevant cross section and can take the value of

            * pp - positive positive non-spin-flip cross section,
            * nn - negative negative non-spin-flip cross section,
            * pn - positive negative spin-flip cross section,
            * np - negative positive spin-flip cross section,
            * p - positive half polarised cross section,
            * n - negitive half polarised cross section,
            * unpol - unpolarised cross section.
    polarisation : turple
        Defines the polarisation direction of the incoming reutron beam
        with respect to the sample reference frame.

    Returns
    -------
    discretisedfield.Field
        Scattering cross section in arbitary units.

    Examples
    --------

    .. plot::
        :context: close-figs

        1. Visualising unpolarised cross section with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import numpy as np
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
        ...                p2=(25e-9, 25e-9, 50e-9),
        ...                cell=(1e-9, 1e-9, 2e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     q = 10e-9
        ...     return (0,
        ...             np.sin(2 * np.pi * x / q),
        ...             np.cos(2 * np.pi * x / q))
        >>> field = df.Field(mesh, nvdim=3, value=v_fun, norm=1e5)
        >>> field.sel('z').mpl()
        >>> cs = mag2exp.sans.cross_section(field, method='unpol',
        ...                                 polarisation=(0, 0, 1))
        >>> cs.sel(k_z=0).real.mpl.scalar()

    .. plot::
        :context: close-figs

        2. Visualising spin-flip cross section with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import numpy as np
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
        ...                p2=(25e-9, 25e-9, 50e-9),
        ...                cell=(1e-9, 1e-9, 2e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     q = 10e-9
        ...     return (0,
        ...             np.sin(2 * np.pi * x / q),
        ...             np.cos(2 * np.pi * x / q))
        >>> field = df.Field(mesh, nvdim=3, value=v_fun, norm=1e5)
        >>> field.sel('z').mpl()
        >>> cs = mag2exp.sans.cross_section(field, method='pn',
        ...                                 polarisation=(0, 0, 1))
        >>> cs.sel(k_z=0).real.mpl.scalar()
    """
    cross_s = _cross_section_matrix(field, polarisation=polarisation)
    # TODO: make more efficient!

    # norm_field = df.Field(field.mesh, nvdim=1, value=(field.norm.array != 0))
    # volume = norm_field.integrate()

    factor = 1  # 8 * np.pi**3 * (2.91e8)**2 / (field.mesh.dV)

    if method in ("polarised_pp", "pp"):
        return factor * cross_s.pp
    elif method in ("polarised_pn", "pn"):
        return factor * cross_s.pn
    elif method in ("polarised_np", "np"):
        return factor * cross_s.np
    elif method in ("polarised_nn", "nn"):
        return factor * cross_s.nn
    elif method in ("half_polarised_p", "p"):
        return factor * cross_s.pp + cross_s.pn
    elif method in ("half_polarised_n", "n"):
        return factor * cross_s.nn + cross_s.np
    elif method in ("unpolarised", "unpol"):
        return 0.5 * factor * (cross_s.pp + cross_s.pn + cross_s.np + cross_s.nn)
    else:
        msg = f"Method {method} is unknown."
        raise ValueError(msg)


def chiral_function(field, /, polarisation=(0, 0, 1)):
    r"""Calculation of the chiral function :math:`-2\pi i \chi`.

    The chiral function can be calculated using

    .. math::
        -2\pi i \chi = \frac{d\sum^{+-}}{d\Omega} - \frac{d\sum^{-+}}{d\Omega}

    where :math:`\frac{d\sum^{+-}}{d\Omega}` and
    :math:`\frac{d\sum^{-+}}{d\Omega}` are the spin-flip cross sections.

    Note that this function returns the quantity :math:`-2\pi i \chi`.

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.
    polarisation : turple
        Defines the polarisation direction of the incoming reutron beam
        with respect to the sample reference frame.

    Returns
    -------
    discretisedfield.Field
        Chiral function :math:`-2\pi i \chi`.

    Examples
    --------

    .. plot::
        :context: close-figs

        1. Visualising the chiral function with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import numpy as np
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -25e-9),
        ...                p2=(25e-9, 25e-9, 25e-9),
        ...                cell=(1e-9, 1e-9, 1e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     q = 10e-9
        ...     return (0,
        ...             np.sin(2 * np.pi * x / q),
        ...             np.cos(2 * np.pi * x / q))
        >>> field = df.Field(mesh, nvdim=3, value=v_fun, norm=1e5)
        >>> field.sel('z').mpl()
        >>> cf = mag2exp.sans.chiral_function(field,
        ...                                   polarisation=(1, 0, 0))
        >>> cf.sel(k_z=0).mpl.scalar()
    """
    cross_s = _cross_section_matrix(field, polarisation=polarisation)
    return cross_s.pn - cross_s.np


def _cross_section_matrix(field, /, polarisation):
    m_fft = field.fftn(norm="ortho")
    m_fft *= field.mesh.dV
    q = df.Field(
        m_fft.mesh,
        nvdim=3,
        value=(
            lambda x: (0, 0, 0) if np.linalg.norm(x) == 0 else x / np.linalg.norm(x)
        ),
    )
    magnetic_interaction = q.cross(m_fft.cross(q))

    # Rotation of Pauli matrices
    initial = (0, 0, 1)
    if initial == polarisation:
        r = Rotation.identity()
    else:
        fixed = np.cross(initial, polarisation)
        r = Rotation.align_vectors([polarisation, fixed], [initial, fixed])[0]
    p_x = [[0, 1], [1, 0]]
    p_y = [[0, -1j], [1j, 0]]
    p_z = [[1, 0], [0, -1]]
    p = np.array([p_x, p_y, p_z])
    p_new = np.einsum("ij,ibc->jbc", r.as_matrix(), p)  # Rotate Pauli matrices

    # Apply function to Pauli
    magnetic_interaction_new = np.einsum(
        "ijkl,lbc->ijkbc", magnetic_interaction.array, p_new
    )
    cs = np.power(np.abs(magnetic_interaction_new), 2)
    return df.Field(
        mesh=m_fft.mesh,
        nvdim=4,
        vdims=["pp", "np", "pn", "nn"],
        value=cs.reshape([*cs.shape[:3], 4]),
        vdim_mapping={},
    )



"""Magnetometry submodule.

Module for calculation of magnetometry based techniques.
"""

import discretisedfield as df
import micromagneticmodel as mm
import numpy as np


def magnetisation(field):
    r"""Calculation of the magnetisation.

    Calculates the magnetisation in the regions where the norm is non-zero.

    Parameters
    ----------
    field : discretisedfield.Field
        Magnetisation field.

    Returns
    -------
    Magnetisation : tuple
        Magnetisation in :math:`\textrm{Am}^{-1}`.

    Examples
    --------

    1.Uniform magnetisation in :math:`z` direction.

    >>> import discretisedfield as df
    >>> import micromagneticmodel as mm
    >>> import mag2exp
    >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
    ...                p2=(25e-9, 25e-9, 50e-9),
    ...                cell=(1e-9, 1e-9, 2e-9))
    >>> field= df.Field(mesh, nvdim=3, value=(0,0,1), norm=1e6)
    >>> mag2exp.magnetometry.magnetisation(field)
    array([      0.,       0., 1000000.])

    2. Spatially dependent magnetisation.

    >>> import discretisedfield as df
    >>> import micromagneticmodel as mm
    >>> import mag2exp
    >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
    ...                p2=(25e-9, 25e-9, 50e-9),
    ...                cell=(1e-9, 1e-9, 2e-9))
    >>> def v_fun(point):
    ...     x, y, z = point
    ...     if x < -2e-9:
    ...         return (0, 0, 1)
    ...     elif x < 2e-9:
    ...         return (0, 1, 0)
    ...     else:
    ...         return (-1, 1, 0)
    >>> field= df.Field(mesh, nvdim=3, value=v_fun, norm=1e6)
    >>> mag2exp.magnetometry.magnetisation(field)
    array([-325269.11934575,  405269.1193456 ,  460000.        ])

    """
    # TODO: Valid volume
    return field.mean() / field.orientation.norm.mean()


def torque(field, H):
    r"""Calculation of the torque.

    The torque is calculated using

    .. math::
        {\bf \tau} = {\bf m} \times {\bf B},

    where :math:`{\bf B}` is the magnetic flux density and :math:`{\bf m}` is
    the magnetic moment.
    The magnetisation :math:`{\bf M}` can be related to the magnetic moment
    using

    .. math::
        {\bf M} = \frac{d{\bf m}}{dV},

    where :math:`dV` is a volume element.

    These equations can be written as

    .. math::
        {\bf \tau} = {\mu_0 {\bf m} \times {\bf H}_{app}},

    where :math:`{\bf H}_{app}` is the applied magnetic field.

    Parameters
    ----------
    field : discretisedfield.Field
        Magnetisation field.

    H : tuple, discretisedfield.Field
        Applied magnetic flux density in :math:`\textrm{Am}^{-1}`.

    Returns
    -------
    tuple
        Torque in :math:`\textrm{Nm}^{-2}`.

    Examples
    --------

    1. Field along magnetisation direction.

    >>> import discretisedfield as df
    >>> import micromagneticmodel as mm
    >>> import mag2exp
    >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
    ...                p2=(25e-9, 25e-9, 50e-9),
    ...                cell=(1e-9, 1e-9, 2e-9))
    >>> system = mm.System(name='Box2')
    >>> system.energy = mm.Zeeman(H=(0, 0, 1e6)) + mm.Demag()
    >>> system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=1e6)
    >>> np.allclose(mag2exp.magnetometry.torque(system.m, system.energy.zeeman.H), 0)
    True

    2. Field perpendicular to magnetisation direction.

    >>> import numpy as np
    >>> import discretisedfield as df
    >>> import micromagneticmodel as mm
    >>> import mag2exp
    >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
    ...                p2=(25e-9, 25e-9, 50e-9),
    ...                cell=(1e-9, 1e-9, 2e-9))
    >>> system = mm.System(name='Box2')
    >>> system.energy = mm.Zeeman(H=(1e6, 0, 0)) + mm.Demag()
    >>> system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=1e6)
    >>> np.allclose(
    ...     mag2exp.magnetometry.torque(system.m, system.energy.zeeman.H),
    ...     (0, mm.consts.mu0*1e12, 0)
    ... )
    True
    """
    # TODO: Valid volume
    # TODO: Field of H
    total_field = mm.consts.mu0 * np.array(H)
    norm_field = df.Field(field.mesh, nvdim=1, value=(field.norm.array != 0))
    volume = norm_field.integrate()
    moment = field * volume[0]
    torque = moment & total_field
    return torque.integrate() / volume**2



"""MOKE submodule.

Module for calculation of magneto-optical Kerr effect
quantities.

**The current version of MOKE microscopy is under development and currently
the angle of incidence is approximated to be the same throughout
the sample. This is not correct but allows the rough behaviour to be seen
and we are looking to correct this in future releases. Please feel free to
contact us or raise an issue if you have any comments on this technique.**
"""

import warnings

import discretisedfield as df
import numpy as np

import mag2exp


def _calculate_A(theta_j, nj, voight, field):
    r"""Calculation of the boundary matrix.

    .. math::
        \begin{equation}
            A_{j} =
            \begin{pmatrix}
            1 & 0 & 1 & 0 \\
            \frac{iQ\alpha^2_{yi}}{2} \left(m_y \frac{1+\alpha^2_{zi}}
            {\alpha_{yi}\alpha_{zi}} - m_z\right) & \alpha_{zi} &
            -\frac{iQ\alpha^2_{yi}}{2} \left(m_y \frac{1+\alpha^2_{zi}}
            {\alpha_{yi}\alpha_{zi}} + m_z\right) & -\alpha_{zi} \\
            -\frac{in_jQ}{2} \left(m_y \alpha_{yi} + m_z \alpha_{zi} \right) &
            -n_j &
            -\frac{in_jQ}{2} \left(m_y \alpha_{yi} - m_z \alpha_{zi} \right) &
            -n_j \\
            n_j\alpha_{zj} & -\frac{in_jQ}{2} \left(m_y \frac{\alpha_{yi}}
            {\alpha_{zi}} - m_z \right) & -n_j \alpha_{zj} &
            \frac{in_jQ}{2} \left(m_y \frac{\alpha_{yi}}
            {\alpha_{zi}} + m_z \right)
            \end{pmatrix}
        \end{equation}

    where :math:`\alpha_{yi}=\sin\theta_j, \alpha_{zi}=\cos\theta_j`,
    :math:`\theta_j` is the complex refractive angle, :math:`Q` is the Voight
    parameter and :math:`n_j` is the refractive index of the :math:`j`th layer.
    The angle is measure with respect to the :math:`z` axis and is calculated
    using Snell's law.
    :math:`m_x`, :math:`m_y` and
    :math:`m_z` are the normalized magnetisation.
    """
    m_arr = field.orientation.array
    a_yj = np.sin(theta_j)
    a_zj = np.cos(theta_j)

    s = np.shape(m_arr)
    mx_arr = m_arr[..., 0].flatten()
    my_arr = m_arr[..., 1].flatten()
    mz_arr = m_arr[..., 2].flatten()

    A = []
    for _mx, my, mz in zip(mx_arr, my_arr, mz_arr):
        A.append(
            [
                [1, 0, 1, 0],
                [
                    (1j * voight)
                    * (a_yj * my * (1 + a_zj**2) / a_zj - mz * a_yj**2)
                    / 2,
                    a_zj,
                    -(1j * voight)
                    / 2
                    * (a_yj * my * (1 + a_zj**2) / (a_zj) + mz * a_yj**2),
                    -a_zj,
                ],
                [
                    -(1j * voight * nj) / 2 * (my * a_yj + mz * a_zj),
                    -nj,
                    -(1j * voight * nj) / 2 * (my * a_yj - mz * a_zj),
                    -nj,
                ],
                [
                    nj * a_zj,
                    -(1j * voight * nj) / 2 * (my * a_yj / a_zj - mz),
                    -nj * a_zj,
                    (1j * voight * nj) / 2 * (my * a_yj / a_zj + mz),
                ],
            ]
        )
    return np.reshape(A, (*s[0:2], 4, 4))


def _calculate_D(theta_j, nj, voight, dj, wavelength, field):
    r"""Calculation of the propagation matrix.

    .. math::
        \begin{equation}
            D_{j} =
            \begin{pmatrix}
            U\cos\delta^i & U\sin\delta^i & 0 & 0 \\
            -U\sin\delta^i & U\cos\delta^i & 0 & 0 \\
            0 & 0 & U^{-1}\cos\delta^r & U^{-1}\sin\delta^r \\
            0 & 0 & -U^{-1}\sin\delta^r & U^{-1}\cos\delta^r
            \end{pmatrix}
        \end{equation}

    where

    .. math::
        \begin{align}
            U &= \exp\left(\frac{-i2\pi n_j \alpha_{zj} d_j}{\lambda} \right)\\
            \delta^i &= -\frac{\pi n_j Q d_j g^i}{\lambda \alpha_{zj}} \\
            \delta^r &= -\frac{\pi n_j Q d_j g^r}{\lambda \alpha_{zj}} \\
            g^i &= m_z \alpha_{zj} + m_y \alpha_{yj} \\
            g^r &= m_z \alpha_{zj} - m_y \alpha_{yj}
        \end{align}

    where :math:`\alpha_{yi}=\sin\theta_j, \alpha_{zi}=\cos\theta_j`,
    :math:`\theta_j` is the complex refractive angle, :math:`Q` is the Voight
    parameter, :math:`n_j` is the refractive index, and :math:`d_j`
    is the thickness of the :math:`j`th layer.
    The angle is measure with respect to the :math:`z` axis and is calculated
    using Snell's law.
    :math:`m_x`, :math:`m_y` and
    :math:`m_z` are the normalized magnetisation.
    :math:`\lambda` is the wavelength of light.
    """
    m_arr = field.orientation.array
    a_yj = np.sin(theta_j)
    a_zj = np.cos(theta_j)

    s = np.shape(m_arr)
    mx_arr = m_arr[..., 0].flatten()
    my_arr = m_arr[..., 1].flatten()
    mz_arr = m_arr[..., 2].flatten()

    D = []
    for _mx, my, mz in zip(mx_arr, my_arr, mz_arr):
        gi = mz * a_zj + my * a_yj
        gr = mz * a_zj - my * a_yj
        di = -np.pi * nj * voight * dj * gi / (wavelength * a_zj)
        dr = -np.pi * nj * voight * dj * gr / (wavelength * a_zj)
        U = np.exp(-2j * np.pi * nj * a_zj * dj / wavelength)

        D.append(
            [
                [U * np.cos(di), U * np.sin(di), 0, 0],
                [-U * np.sin(di), U * np.cos(di), 0, 0],
                [0, 0, np.cos(dr) / U, np.sin(dr) / U],
                [0, 0, -np.sin(dr) / U, np.cos(dr)],
            ]
        )
    return np.reshape(D, (*s[0:2], 4, 4))


def _angle_snell(theta0, n0, n1):
    r"""Calculation of the new angle from Snell's law.

    Snell's is

    .. math::
        n_0 \sin\theta_0 = n_1 \sin\theta_1

    where :math:`n_0` and :math:`n_1` are the refractive indexes of layer 0
    and 1 respectively. :math:`\theta_0` and :math:`\theta_1` are the angles
    of light in layer 0 and 1, with respect to the surface normal.
    """
    return np.arcsin((n0 * np.sin(theta0)) / n1)


def _calculate_M(field, theta_0, n_0, voight, wavelength):
    r"""Product matrix.

    The product matrix is the matrix used to describe light propagation in
    a magnetic multilayer system. In this function, it has the form

    .. math::
        \begin{equation}
            M = A_f^{-1} \prod_{j} A_{j} D_j A_j^{-1} A_f,
        \end{equation}

    where :math:`A_j` is the boundary matrix for the :math:`j`th layer,
    :math:`D_j` is the propagation matrix for the :math:`j`th layer,
    and :math:`A_f` is the boundary layer matrix for free space.
    """
    # Free space
    A = _calculate_A(theta_0, 1, 0, field.sel("z"))
    M = np.linalg.inv(A)

    # Sample
    z_min = field.mesh.index2point((0, 0, 0))[2]
    z_max = field.mesh.index2point(np.subtract(field.mesh.n, 1))[2]
    z_step = field.mesh.cell[2]
    values = np.arange(z_min, z_max + 1e-20, z_step)

    theta = _angle_snell(theta_0, 1, n_0)
    for z in values:  # Think about direction of loop
        A = _calculate_A(theta, n_0, voight, field.sel(z=z))
        D = _calculate_D(theta, n_0, voight, field.mesh.dz, wavelength, field.sel(z=z))
        M = np.matmul(M, A)
        M = np.matmul(M, D)
        M = np.matmul(M, np.linalg.inv(A))

    # Free space
    theta = theta_0
    A = _calculate_A(theta, 1, 0, field.sel("z"))
    M = np.matmul(M, A)
    return M


def _M_to_r(M):
    r"""Product matrix to reflection matrix."""
    G_matrix = M[:, :, 0:2, 0:2]
    I_matrix = M[:, :, 2:4, 0:2]
    G_inv = np.linalg.inv(G_matrix)
    return np.matmul(I_matrix, G_inv)


def _M_to_t(M):
    r"""Product matrix to transmission matrix."""
    G_matrix = M[:, :, 0:2, 0:2]
    G_inv = np.linalg.inv(G_matrix)
    return G_inv


def intensity(field, theta, n, voight, wavelength, E_i, mode="reflection", fwhm=None):
    r"""MOKE intensity.

    **The current version of MOKE microscopy is under development and currently
    the angle of incidence is approximated to be the same throughout
    the sample. This is not correct but allows the rough behaviour to be seen
    and we are looking to correct this in future releases. Please feel free to
    contact us or raise an issue if you have any comments on this technique.**

    Calculation of the intensity of the magneto-optical Kerr effect from an
    incident wave described by

    .. math::
        \begin{pmatrix}
            E^i_s \\ E^i_p ,
        \end{pmatrix}

    where :math:`E^i_s` and :math:`E^i_p` are the electric field components of
    incident linear s and p polarisation modes.

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.
    theta : float
        Defines the and of incidence of the beam relative to the surface normal
        in the :math:`yz` plane
        with respect to the sample reference frame.
    n_0 : float
        Complex refractive index.
    voight : float
        Complex Voight parameter.
    wavelength : float
        Wavelength of the incident beam.
    E_i : list
        Incident electric field vector with components of linear
        s and p polarisation modes.
    mode : str
        Reflection or transmission mode.
    fwhm : array_like, optional
        If specified, convolutes the output image with a 2 Dimensional Gaussian
        with the full width half maximum (fwhm) specified.

    Returns
    -------
    discretisedfield.Field
        Intensity of the MOKE image.

    Examples
    --------

    .. plot::

        1. Visualising the MOKE with ``matplotlib``.

        >>> import warnings
        >>> warnings.filterwarnings("ignore", "This technique is currently under")
        ...
        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import numpy as np
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -25e-9),
        ...                p2=(25e-9, 25e-9, 25e-9),
        ...                cell=(1e-9, 1e-9, 1e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     q = 10e-9
        ...     return (0,
        ...             np.sin(2 * np.pi * x / q),
        ...             np.cos(2 * np.pi * x / q))
        >>> field = df.Field(mesh, nvdim=3, value=v_fun, norm=1e5)
        >>> E_i = [1, 1j]
        >>> intensity = mag2exp.moke.intensity(field, 0, 2, 1, 600e-9, E_i,
        ...                                    mode='reflection')
        >>> intensity.mpl()

    .. seealso::

        :py:func:`~mag2exp.moke.e_field`

    """
    E_f = e_field(field, theta, n, voight, wavelength, E_i, mode=mode)
    intensity = abs(E_f.s) ** 2 + abs(E_f.p) ** 2
    if fwhm is not None:
        intensity = mag2exp.util.gaussian_filter(intensity, fwhm=fwhm)
    return intensity


def kerr_angle(field, theta, n_0, voight, wavelength):
    r"""Kerr angle.

    **The current version of MOKE microscopy is under development and currently
    the angle of incidence is approximated to be the same throughout
    the sample. This is not correct but allows the rough behaviour to be seen
    and we are looking to correct this in future releases. Please feel free to
    contact us or raise an issue if you have any comments on this technique.**

    The Kerr rotation is calculated from

    .. math::
        \begin{align}
            \Phi_s &= \phi_s' + i \phi_s'' \\
            &= \frac{r_{ps}}{r_{ss}} \\
            \Phi_p &= \phi_p' + i \phi_p'' \\
            &= -\text{Real}\left(\frac{r_{sp}}{r_{pp}}\right) +
            \text{Imag}\left(\frac{r_{sp}}{r_{pp}}\right) i
        \end{align}

    where :math:`\phi'` is the Kerr rotation and
    :math:`\phi''` is the ellipticity for the s and p polarisations.
    :math:`r` are the components of the rotation matrix.

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.
    theta : float
        Defines the and of incidence of the beam relative to the surface normal
        in the :math:`yz` plane
        with respect to the sample reference frame.
    n_0 : float
        Complex refractive index.
    voight : float
        Complex Voight parameter.
    wavelength : float
        Wavelength of the incident beam.
    fwhm : array_like, optional
        If specified, convolutes the output image with a 2 Dimensional Gaussian
        with the full width half maximum (fwhm) specified.

    Returns
    -------
    discretisedfield.Field
        Kerr angle of the MOKE.

    Examples
    --------

    .. plot::

        1. Visualising the MOKE with ``matplotlib``.

        >>> import warnings
        >>> warnings.filterwarnings("ignore", "This technique is currently under")
        ...
        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import numpy as np
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -25e-9),
        ...                p2=(25e-9, 25e-9, 25e-9),
        ...                cell=(1e-9, 1e-9, 1e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     q = 10e-9
        ...     return (0,
        ...             np.sin(2 * np.pi * x / q),
        ...             np.cos(2 * np.pi * x / q))
        >>> field = df.Field(mesh, nvdim=3, value=v_fun, norm=1e5)
        >>> E_i = [1, 1j]
        >>> angle = mag2exp.moke.kerr_angle(field, 0, 2, 1, 600e-9)
        >>> angle.s.real.mpl()
    """
    warnings.simplefilter("always", lineno=367)
    warnings.warn(
        "This technique is currently under development so results"
        " may not be accurate."
        " Please see the documentation for further details.",
        stacklevel=2,
    )
    M = _calculate_M(field, theta, n_0, voight, wavelength)
    m = _M_to_r(M)

    k_s = m[..., 1, 0] / m[..., 0, 0]
    k_p = m[..., 0, 1] / m[..., 1, 1]
    k_p = -k_p.real + 1j * k_p.imag

    k_a = np.stack([k_s, k_p], axis=-1)

    angle = df.Field(
        mesh=field.sel("z").mesh,
        nvdim=2,
        value=k_a,
        vdims=["s", "p"],
    )
    return angle


def e_field(field, theta, n_0, voight, wavelength, E_i, mode="reflection"):
    r"""Electric field.

    **The current version of MOKE microscopy is under development and currently
    the angle of incidence is approximated to be the same throughout
    the sample. This is not correct but allows the rough behaviour to be seen
    and we are looking to correct this in future releases. Please feel free to
    contact us or raise an issue if you have any comments on this technique.**

    Calculate the reflected or transmitted electric field using

    .. math::
        \begin{equation}
            \begin{pmatrix} E^r_s \\ E^r_p \end{pmatrix} =
            \begin{pmatrix} r_{ss} & r_{sp} \\ r_{ps} & r_{pp} \end{pmatrix}
            \begin{pmatrix} E^i_s \\ E^i_p \end{pmatrix}
        \end{equation},

    where E^i_s and E^i_p are the electric field components of incident linear
    s and p polarisation modes.

    The reflection and transmission matrices are calculated from the product
    matrix.

    The product matrix is the matrix used to describe light propagation in
    a magnetic multilayer system. In this function, it has the form

    .. math::
        \begin{equation}
            M = A_f^{-1} \prod_{j} A_{j} D_j A_j^{-1} A_f,
        \end{equation}

    where :math:`A_j` is the boundary matrix for the :math:`j`th layer,
    :math:`D_j` is the propagation matrix for the :math:`j`th layer,
    and :math:`A_f` is the boundary layer matrix for free space.

    In the system described here, we assume the magnetic structure is
    surrounded by free space.

    The beam in incident in the :math:`yz` plane with :math:`\theta` the angle
    from the normal.


    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.
    theta : float
        Defines the and of incidence of the beam relative to the surface normal
        in the :math:`yz` plane
        with respect to the sample reference frame.
    n_0 : float
        Complex refractive index.
    voight : float
        Complex Voight parameter.
    wavelength : float
        Wavelength of the incident beam.
    E_i : list
        Incident electric field vector with components of linear
        s and p polarisation modes.
    mode : str
        Reflection or transmission mode.

    Returns
    -------
    discretisedfield.Field
        Electric field of reflected and transmitted wave with components ``s``
        and ``p`` representing the polarisation.

    Examples
    --------

    .. plot::

        1. Calculating the reflected electric field from a circularly polarised
        wave.

        >>> import warnings
        >>> warnings.filterwarnings("ignore", "This technique is currently under")
        ...
        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import numpy as np
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -25e-9),
        ...                p2=(25e-9, 25e-9, 25e-9),
        ...                cell=(1e-9, 1e-9, 1e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     q = 10e-9
        ...     return (0,
        ...             np.sin(2 * np.pi * x / q),
        ...             np.cos(2 * np.pi * x / q))
        >>> field = df.Field(mesh, nvdim=3, value=v_fun, norm=1e5)
        >>> E_i = [1, 1j]
        >>> E_f = mag2exp.moke.e_field(field, 0, 2, 1, 600e-9, E_i,
        ...                            mode='reflection')
    """
    warnings.simplefilter("always", lineno=489)
    warnings.warn(
        "This technique is currently under development so results"
        " may not be accurate."
        " Please see the documentation for further details.",
        stacklevel=2,
    )
    M = _calculate_M(field, theta, n_0, voight, wavelength)
    if mode in ("reflection", "r"):
        m = _M_to_r(M)
    elif mode in ("transmission", "t"):
        m = _M_to_t(M)
    else:
        msg = f"Mode {mode} is unknown."
        raise ValueError(msg)

    E_f = np.matmul(m, E_i)
    E_f_field = df.Field(
        mesh=field.sel("z").mesh,
        nvdim=2,
        value=E_f,
        vdims=["s", "p"],
    )

    return E_f_field



import discretisedfield as df
import pytest

import mag2exp


def test_util_gaussian_filter_dim():
    fwhm = (1e-9, 1e-9, 1e-9)
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    with pytest.raises(RuntimeError):
        mag2exp.util.gaussian_filter(field, fwhm)
    new_field = mag2exp.util.gaussian_filter(field.z, fwhm)
    assert new_field.nvdim == 1


def test_util_gaussian_filter_plane():
    fwhm = (1e-9, 1e-9, 1e-9)
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    new_field = mag2exp.util.gaussian_filter(field.z.sel("z"), fwhm)
    assert new_field.mesh.region.ndim == 2


def test_util_gaussian_filter_fwhm():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    fwhm = (1e-9, 1e-9)
    mag2exp.util.gaussian_filter(field.z.sel("z"), fwhm)
    fwhm = (1e-9, 1e-9, 1e-9)
    mag2exp.util.gaussian_filter(field.z, fwhm)
    fwhm = (1e-9, 1e-9, 1e-9, 1e-9)
    mag2exp.util.gaussian_filter(field.z, fwhm)


def test_util_calculate_demag():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.util.calculate_demag_field(field)



"""LTEM submodule.

Module for calculation of Lorentz Transmission Electron Microscopy related
quantities.
"""

import warnings

import discretisedfield as df
import micromagneticmodel as mm
import numpy as np
from scipy import constants


def phase(field, /, kcx=0.1, kcy=0.1):
    r"""Calculation of the magnetic phase shift experienced by the electrons.

    The Fourier transform of the magnetic phase shift is calculated using

    .. math::

        \widetilde{\phi}_m (k_x,k_y) = \frac {i e \mu_0 k_\perp^2}{h}
        \frac{\left[ \widetilde{\bf M}_I(k_x,k_y) \times
        {\bf k}_\perp \right] _z}{\left( k_\perp^2 + k_c^2 \right)^2},

    where :math:`{\mathbf{M}}_I` is the integrated magnetisation along the path
    of the electron beam. Here we define the electron beam to be
    propagating in the :math:`z` direction.
    :math:`\mu_0` is the vacuum permeability, and :math:`k` is the k-vector in
    Fourier space.
    :math:`k_c` is the radius of the filter and can be written in 2-Dimensions
    as to

    .. math::

        k_c^2 = \left(k_{cx} dk_x\right) ^2 +  \left(k_{cy} dk_y \right) ^2,

    where :math:`dk_x` and :math:`dk_y` are the resolution in Fourier space for
    the :math:`x` and :math:`y` directions respectively. :math:`k_{cx}` and
    :math:`k_{cy}` are the radii of the filter in each direction in units of
    cells.


    Parameters
    ----------
    field : discretisedfield.Field
        Magnetisation field.
    kcx : numbers.Real, optional
        Tikhonov filter radius in :math:`x` in units of cells.
    kcy : numbers.Real, optional
        Tikhonov filter radius in :math:`y` in units of cells.

    Returns
    -------
    discretisedfield.Field
        Phase in real space.
    discretisedfield.Field
        Phase in Fourier space.

    Examples
    --------

    1. Uniform out-of-plane field.

    >>> import discretisedfield as df
    >>> import mag2exp
    >>> mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 1), cell=(1, 1, 1))
    >>> field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    >>> phase, ft_phase = mag2exp.ltem.phase(field)
    >>> phase.mean()
    array([0.])

    .. plot::
        :context: close-figs

        2. Visualising the phase using ``matplotlib``.

        >>> import discretisedfield as df
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-5, -4, -1), p2=(5, 4, 1), cell=(2, 1, 0.5))
        ...
        >>> def value_fun(point):
        ...     x, y, z = point
        ...     if x > 0:
        ...         return (0, -1, 0)
        ...     else:
        ...         return (0, 1, 0)
        ...
        >>> field = df.Field(mesh, nvdim=3, value=value_fun)
        >>> phase, ft_phase = mag2exp.ltem.phase(field)
        >>> phase.mpl.scalar()

    """
    # More readable notation, direction arg will be removed soon.
    m_int = field.integrate(direction="z")
    m_ft = m_int.fftn(norm="ortho")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        k = df.Field(m_ft.mesh, nvdim=3, value=lambda x: (x[0], x[1], 0))
    denom = (k.x**2 + k.y**2) / (
        k.x**2 + k.y**2 + k.mesh.dk_x**2 * kcx**2 + k.mesh.dk_y**2 * kcy**2
    ) ** 2

    prefactor = 1j * mm.consts.e * mm.consts.mu0 / mm.consts.h
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ft_phase = (m_ft & k).ft_z * denom * prefactor
    phase = ft_phase.ifftn(norm="ortho").real
    phase.mesh.translate(field.mesh.region.center[:2], inplace=True)
    return phase, ft_phase


def defocus_image(
    phase, /, cs=0, df_length=0.2e-3, voltage=None, wavelength=None, ca2=0, phia2=0
):
    r"""Calculating the defocused image.

    The wavefunction of the electrons is created from the magnetic phase shift
    :math:`\phi_m`

    .. math::

        \psi_0 = e^{i\phi_m},

    and propagated through the electron microscope to the image plane by use of
    the Contrast Transfer Function :math:`T`.
    The wavefunction at a defocus length :math:`\Delta f` in Fourier space is
    given by

    .. math::

        \widetilde{\psi}_{\Delta f} = \widetilde{\psi}_0 e^{2 i \pi \lambda k^2
        (-\frac{1}{2} \Delta f + \frac{1}{4} C_s \lambda^2 k^2)},

    where :math:`\lambda` is the relativistic wavelength of the electrons,
    :math:`C_s` is the spherical aberration coefficient of the microscope, and
    :math:`k` is the wavevector in Fourier space.

    The intensity of an image at a specific defocus is given by

    .. math::

        I_{\Delta f} = \left\vert \psi_{\Delta f} \right\vert^2.

    In focus :math:`I=\left\vert\psi_0\right\vert^2=1`

    Either electron ``wavelength`` or acceleration voltage ``U`` must be
    specified.

    Parameters
    ----------
    phase : discretisedfield.Field
        Phase of the electrons from LTEM.
    cs : numbers.Real, optional
        Spherical aberration coefficient.
    df_length : numbers.Real, optional
        Defocus length in m.
    voltage : numbers.Real, optional
        Accelerating voltage of electrons in V.
    wavelength : numbers.Real, optional
        Relativistic wavelength of the electrons in m.

    Returns
    -------
    discretisedfield.Field
        Intensity at specified defocus.

    Examples
    --------

    1. Zero defocus.

    >>> import discretisedfield as df
    >>> import mag2exp
    >>> import numpy as np
    >>> mesh = df.Mesh(p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9),
    ...                    cell=(2e-9, 1e-9, 0.5e-9))
    ...
    >>> def value_fun(point):
    ...     x, y, z = point
    ...     if x > 0:
    ...         return (0,-1, 0)
    ...     else:
    ...         return (0, 1, 0)
    ...
    >>> field = df.Field(mesh, nvdim=3, value=value_fun)
    >>> phase, ft_phase = mag2exp.ltem.phase(field)
    >>> df_img = mag2exp.ltem.defocus_image(phase, cs=0, df_length=0,
    ...                                     voltage=300e3)
    >>> np.allclose(df_img.array, [1])
    True

    .. plot::
        :context: close-figs

        2. Visualising the phase using ``matplotlib``.

        >>> import discretisedfield as df
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9),
        ...                    cell=(2e-9, 1e-9, 0.5e-9))
        ...
        >>> def value_fun(point):
        ...     x, y, z = point
        ...     if x > 0:
        ...         return (0, -1, 0)
        ...     else:
        ...         return (0, 1, 0)
        ...
        >>> field = df.Field(mesh, nvdim=3, value=value_fun)
        >>> phase, ft_phase = mag2exp.ltem.phase(field)
        >>> df_img = mag2exp.ltem.defocus_image(phase, cs=8000,
        ...                                     df_length=0.2e-3,
        ...                                     voltage=300e3)
        >>> df_img.mpl.scalar()

    .. seealso::

        :py:func:`~mag2exp.ltem.phase`
        :py:func:`~mag2exp.ltem.relativistic_wavelength`

    """
    ft_wavefn = np.exp(phase * 1j).fftn(norm="ortho")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        k = df.Field(
            ft_wavefn.mesh,
            nvdim=3,
            value=lambda x: (x[0], x[1], 0),
            dtype=np.complex128,
        )
        ksquare = k.x**2 + k.y**2
        phi = df.Field(phase.mesh, nvdim=1, value=lambda x: np.arctan2(x[1], x[0]))

    if wavelength is None:
        if voltage is None:
            msg = "Either `wavelength` or acceleration `voltage` needs to be specified."
            raise RuntimeError(msg)
        wavelength = relativistic_wavelength(voltage)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        cts = (
            -np.pi * df_length * wavelength * ksquare
            + 0.5 * np.pi * cs * wavelength**3 * ksquare**2
            + np.pi * ca2 * wavelength * ksquare * np.sin(2 * (phi.array - phia2))
        )
        exp = np.exp(-1j * cts)
    ft_def_wf_cts = ft_wavefn * exp
    def_wf_cts = ft_def_wf_cts.ifftn(norm="ortho")
    intensity_cts = def_wf_cts.conjugate * def_wf_cts
    intensity_cts.mesh.translate(phase.mesh.region.center, inplace=True)
    return intensity_cts.real


def integrated_magnetic_flux_density(phase):
    r"""Calculate the integrated magnetic flux density.

    This calculates the magnetic flux density integrated along the beam
    direction given by

    .. math::
        \int_{0}^{t} {\bf B}_\perp dz = \frac{\Phi_0}{\pi}\left(
        \begin{array}{c}
         -\partial/\partial y \\
          \partial/\partial x
        \end{array} \right).


    This quantity is most closely related to values obtained from electron
    holography.

    Parameters
    ----------
    phase : discretisedfield.Field
        Phase of the electrons from LTEM.

    Returns
    -------
    discretisedfield.Field
        Integrated magnetic flux density.

    Example
    -------
    .. plot::
        :context: close-figs

        1. Visualising the phase using ``matplotlib``.

        >>> import discretisedfield as df
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9),
        ...                    cell=(2e-9, 1e-9, 0.5e-9))
        ...
        >>> def value_fun(point):
        ...     x, y, z = point
        ...     if x > 0:
        ...         return (0, -1, 0)
        ...     else:
        ...         return (0, 1, 0)
        ...
        >>> field = df.Field(mesh, nvdim=3, value=value_fun)
        >>> phase, ft_phase = mag2exp.ltem.phase(field)
        >>> df_img = mag2exp.ltem.defocus_image(phase, cs=8000,
        ...                                     df_length=0.2e-3,
        ...                                     voltage=300e3)
        >>> imf = mag2exp.ltem.integrated_magnetic_flux_density(phase)

    """
    imfd = -phase.real.diff("y") << phase.real.diff("x")
    return mm.consts.hbar / mm.consts.e * imfd


def relativistic_wavelength(voltage):
    r"""Relativistic wavelength of an electron accelerated using a voltage U.

    The wavelength is calculated using

    .. math::
        \lambda = \frac{h}{
        \sqrt{2 m_e e U \left(1 + \frac{e U}{2 m_e c^2}\right)}}

    where :math:`m_e` and :math:`e` are the mass and charge of an electron
    respectively, :math:`c` is the speed of light, and :math:`U` is the
    accelerating voltage in V.


    Parameters
    ----------
    voltage : numbers.Real
        Accelerating voltage in V.

    Returns
    -------
    numbers.Real
        Wavelength in m.

    Example
    -------
    1. Accelerating using 300 kV.

    >>> import mag2exp
    >>> mag2exp.ltem.relativistic_wavelength(300e3)
    1.9687...e-12

    """
    return (
        constants.h
        / np.sqrt(
            2
            * constants.m_e
            * voltage
            * constants.e
            * (1 + constants.e * voltage / (2 * constants.m_e * constants.c**2))
        )
    ).item()



"""X-ray submodule.

Module for calculation of x-ray based quantities.
"""

import mag2exp


def holography(field, /, fwhm=None):
    r"""Calculation of the pattern obtained from X-ray holography.

    X-ray holography uses magnetic circular dichroism to measure the magnetic
    field parallel to the propagation direction of the light. Here, we define
    the experimental reference frame with the light propagating along :math:`z`
    direction. The results of X-ray holographic imaging is the real space
    integral of the magnetisation along the beam direction.

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.
    fwhm : array_like, optional
        If specified, convolutes the output image with a 2 Dimensional Gaussian
        with the full width half maximum (fwhm) specified.

    Returns
    -------
    discretisedfield.Field
        X-ray holographic image.

    Examples
    --------

    .. plot::
        :context: close-figs

        1. Visualising X-ray holography with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-5e-9, -4e-9, 0), p2=(5e-9, 4e-9, 2e-9),
        ...                cell=(1e-9, 1e-9, 1e-9))
        >>> def value_fun(point):
        ...     x, y, z = point
        ...     if x < -2e-9:
        ...         return (0, 0, 1)
        ...     elif x < 2e-9:
        ...         return (0, 1, 0)
        ...     else:
        ...         return (0, 0, -1)
        >>> field = df.Field(mesh, nvdim=3, value=value_fun, norm=0.3e6)
        >>> xrh = mag2exp.x_ray.holography(field)
        >>> xrh.mpl.scalar()


    .. plot::
        :context: close-figs

        1. Visualising X-ray holography with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-5e-9, -4e-9, 0), p2=(5e-9, 4e-9, 2e-9),
        ...                cell=(1e-9, 1e-9, 1e-9))
        >>> def value_fun(point):
        ...     x, y, z = point
        ...     if x < -2e-9:
        ...         return (0, 0, 1)
        ...     elif x < 2e-9:
        ...         return (0, 1, 0)
        ...     else:
        ...         return (0, 0, -1)
        >>> field = df.Field(mesh, nvdim=3, value=value_fun, norm=0.3e6)
        >>> xrh2 = mag2exp.x_ray.holography(field, fwhm=(2e-9,2e-9))
        >>> xrh2.mpl.scalar()
    """
    # Direction arg will be removed soon.
    magnetisation = field.z.integrate(direction="z")
    if fwhm is not None:
        magnetisation = mag2exp.util.gaussian_filter(magnetisation, fwhm=fwhm)
    return magnetisation


def saxs(field):
    r""" Calculation of the small angle X-ray scattering pattern.

    Small angle X-ray scattering uses magnetic circular dichroism to measure
    the magnetic
    field parallel to the propagation direction of the light. Here, we define
    the experimental reference frame with the light propagating along :math:`z`
    direction. The intensity of scattering can be calculated
    from the Fourier transform of the real space integral of the magnetisation
    along the beam direction multiplied by its complex conjugate.

    .. math::
        M_I = \int M dz, \\
        I \propto \frac{1}{V}\left\vert \widetilde{M_I} \right\vert ^2,

    where :math:`V` is the scattering volume.

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.

    Returns
    -------
    discretisedfield.Field
        Small angle X-ray scattering intensity, arbitrary units.

    Examples
    --------
    .. plot::
        :context: close-figs

        1. Visualising the scattering with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import numpy as np
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(0, 0, 0), p2=(50e-9, 50e-9, 1e-9),
        ...                cell=(1e-9, 1e-9, 1e-9))
        >>> def value_fun(point):
        ...     x, y, z = point
        ...     qx = 10e-9
        ...     return (np.sin(2 * np.pi * x/ qx),
        ...             0,
        ...             np.cos(2 * np.pi * x/ qx))
        >>> field = df.Field(mesh, nvdim=3, value=value_fun, norm=0.3e6)
        >>> xrs = mag2exp.x_ray.saxs(field)
        >>> xrs.mpl.scalar()

    """
    m_fft = field.fftn(norm="ortho").ft_z.sel(k_z=0)
    return abs(m_fft) ** 2



import discretisedfield as df
import numpy as np
import pytest

import mag2exp

THRESH = 1e-64


def test_sans_analytical_parallel_bloch():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 1e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [0, Ms * np.sin(2 * np.pi * x / qx), Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)

    sans = mag2exp.sans.cross_section(m, method="unpol").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, method="pp").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, method="nn").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, method="pn").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, method="np").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2


def test_sans_analytical_parallel_neel():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 1e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [Ms * np.sin(2 * np.pi * x / qx), 0, Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)

    sans = mag2exp.sans.cross_section(m, method="unpol").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, method="pp").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, method="nn").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, method="pn").sel(k_z=0)
    assert np.isclose(sans.array, 0).all()

    sans = mag2exp.sans.cross_section(m, method="np").sel(k_z=0)
    assert np.isclose(sans.array, 0).all()


def test_sans_analytical_perpendicular_neel():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 1e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [Ms * np.sin(2 * np.pi * x / qx), 0, Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="unpol").sel(
        k_z=0
    )
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="nn").sel(k_z=0)
    assert np.isclose(sans.array, 0).all()

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="pp").sel(k_z=0)
    assert np.isclose(sans.array, 0).all()

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="pn").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="np").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2


def test_sans_analytical_perpendicular_bloch():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 1e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [0, Ms * np.sin(2 * np.pi * x / qx), Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="unpol").sel(
        k_z=0
    )
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 2

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="nn").sel(k_z=0)
    assert np.isclose(sans.array, 0).all()

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="pp").sel(k_z=0)
    assert np.isclose(sans.array, 0).all()

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="pn").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(q, 1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 1

    sans = mag2exp.sans.cross_section(m, polarisation=[1, 0, 0], method="np").sel(k_z=0)
    idx = np.unravel_index(sans.array.argmax(), sans.array.shape)[0:2]
    q = sans.mesh.index2point(idx)[0]
    assert np.isclose(q, -1 / qx)
    peaks = (sans.array > THRESH).sum()
    assert peaks == 1


def test_sans_chiral_parallel():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 1e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [Ms * np.sin(2 * np.pi * x / qx), 0, Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    cf = mag2exp.sans.chiral_function(m).sel(k_z=0)
    assert np.isclose(cf.array, 0).all()

    def m_fun(pos):
        x, y, z = pos
        return [0, Ms * np.sin(2 * np.pi * x / qx), Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    cf = mag2exp.sans.chiral_function(m).sel(k_z=0)
    assert np.isclose(cf.array, 0).all()


def test_sans_chiral_perpendicular():
    region = df.Region(p1=(-50e-9, -50e-9, -50e-9), p2=(50e-9, 50e-9, 50e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 5e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [0, Ms * np.sin(2 * np.pi * x / qx), Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    cf = mag2exp.sans.chiral_function(m, polarisation=[1, 0, 0]).sel(k_z=0)
    idx = np.unravel_index(cf.array.argmax(), cf.array.shape)[0:2]
    q = cf.mesh.index2point(idx)[0]
    assert np.isclose(q, 1 / qx)
    peaks = (cf.array > THRESH).sum()
    assert peaks == 1
    idx = np.unravel_index(cf.array.argmin(), cf.array.shape)[0:2]
    q = cf.mesh.index2point(idx)[0]
    assert np.isclose(q, -1 / qx)
    peaks = (cf.array < -THRESH).sum()
    assert peaks == 1

    def m_fun(pos):
        x, y, z = pos
        return [Ms * np.sin(2 * np.pi * x / qx), 0, Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    cf = mag2exp.sans.chiral_function(m, polarisation=[1, 0, 0]).sel(k_z=0)
    assert np.isclose(cf.array, 0).all()


def test_sans_normalisation():
    Ms = 1.1e6

    def m_fun(pos):
        x, y, z = pos
        qx = 20e-9
        return (0, np.sin(2 * np.pi * x / qx), np.cos(2 * np.pi * x / qx))

    region = df.Region(p1=(0, 0, 0), p2=(80e-9, 80e-9, 80e-9))
    mesh = df.Mesh(region=region, cell=(4e-9, 4e-9, 4e-9))
    field1 = df.Field(mesh, nvdim=3, value=m_fun, norm=Ms)
    sans1 = mag2exp.sans.cross_section(field1, method="unpol")
    m1 = abs(sans1.array).max()

    region2 = df.Region(p1=(0, 0, 0), p2=(80e-9, 80e-9, 80e-9))
    mesh2 = df.Mesh(region=region2, cell=(2e-9, 2e-9, 2e-9))
    field2 = df.Field(mesh2, nvdim=3, value=m_fun, norm=Ms)
    sans2 = mag2exp.sans.cross_section(field2, method="unpol")
    m2 = abs(sans2.array).max()

    region = df.Region(p1=(0, 0, 0), p2=(100e-9, 100e-9, 100e-9))
    mesh = df.Mesh(region=region, cell=(4e-9, 4e-9, 4e-9))
    field3 = df.Field(mesh, nvdim=3, value=m_fun, norm=Ms)
    sans3 = mag2exp.sans.cross_section(field3, method="unpol")
    m3 = abs(sans3.array).max()

    assert np.isclose(m2 / m1, ((2 / 4) ** 3))
    assert np.isclose(m3 / m1, (100 / 80) ** 3)


def test_sans_cross_section_methods():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 3e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [0, Ms * np.sin(2 * np.pi * x / qx), Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)

    mag2exp.sans.cross_section(m, method="unpol")
    mag2exp.sans.cross_section(m, method="p")
    mag2exp.sans.cross_section(m, method="n")
    mag2exp.sans.cross_section(m, method="np")
    mag2exp.sans.cross_section(m, method="pn")
    mag2exp.sans.cross_section(m, method="nn")
    mag2exp.sans.cross_section(m, method="pp")
    with pytest.raises(ValueError):
        mag2exp.sans.cross_section(m, method="blah")



# import pytest
import discretisedfield as df
import numpy as np

import mag2exp


def test_xray_holography_inplane():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(1, 1, 0), norm=384e3)
    xrh = mag2exp.x_ray.holography(field)
    assert (xrh.array == 0).all()


def test_xray_holography_outofplane():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=384e3)
    xrh = mag2exp.x_ray.holography(field)
    assert (xrh.array != 0).all()


def test_xray_holography_filter():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=384e3)
    xrh = mag2exp.x_ray.holography(field, [2e-9, 2e-9])
    assert (xrh.array != 0).all()


def test_xray_holography_analytical():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 0.3e-9))
    Ms = 1.1e6

    def m_fun(pos):
        x, y, z = pos
        qx = 30e-9
        qz = 20e-9
        return [0, 0, Ms * np.cos(2 * np.pi * x / qx) * np.sin(2 * np.pi * z / qz)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    holo = mag2exp.x_ray.holography(m)

    def a_fun(pos):
        x, y = pos
        qx = 30e-9
        qz = 20e-9
        analytical = (
            -qz
            / (2 * np.pi)
            * Ms
            * np.cos(2 * np.pi * x / qx)
            * (np.cos(2 * np.pi * 30e-9 / qz) - 1)
        )
        return analytical

    an_holo = df.Field(holo.mesh, nvdim=1, value=a_fun)
    assert np.isclose(holo.array, an_holo.array, rtol=1e-3).all()


def test_xray_saxs():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=384e3)
    xrh = mag2exp.x_ray.saxs(field)
    assert (xrh.array != 0).any()
    assert (np.isreal(xrh.array)).all()


def test_xray_saxs_inplane():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(1, 1, 0), norm=384e3)
    xrh = mag2exp.x_ray.saxs(field)
    assert (xrh.array == 0).all()


def test_xray_saxs_analytical():
    region = df.Region(p1=(-50e-9, -100e-9, 0), p2=(50e-9, 100e-9, 30e-9))
    mesh = df.Mesh(region=region, cell=(5e-9, 5e-9, 1e-9))
    Ms = 1.1e6
    qx = 20e-9

    def m_fun(pos):
        x, y, z = pos
        return [0, 0, Ms * np.cos(2 * np.pi * x / qx)]

    m = df.Field(mesh, nvdim=3, value=m_fun)
    saxs = mag2exp.x_ray.saxs(m)
    idx = np.unravel_index(saxs.array.argmax(), saxs.array.shape)[0:2]
    q = saxs.mesh.index2point(idx)[0]
    assert np.isclose(abs(q), 1 / qx)
    peaks = (saxs.array > 10).sum()
    assert peaks == 2


def test_xray_saxs_normalisation():
    Ms = 1.1e6

    def m_fun(pos):
        x, y, z = pos
        qx = 25e-9
        return (0, np.sin(2 * np.pi * x / qx), np.cos(2 * np.pi * x / qx))

    region = df.Region(p1=(0, 0, 0), p2=(80e-9, 80e-9, 80e-9))
    mesh = df.Mesh(region=region, cell=(4e-9, 4e-9, 4e-9))
    field1 = df.Field(mesh, nvdim=3, value=m_fun, norm=Ms)
    saxs1 = mag2exp.x_ray.saxs(field1)
    m1 = abs(saxs1.array).max()

    region2 = df.Region(p1=(0, 0, 0), p2=(80e-9, 80e-9, 80e-9))
    mesh2 = df.Mesh(region=region2, cell=(2e-9, 2e-9, 2e-9))
    field2 = df.Field(mesh2, nvdim=3, value=m_fun, norm=Ms)
    saxs2 = mag2exp.x_ray.saxs(field2)
    m2 = abs(saxs2.array).max()

    region = df.Region(p1=(0, 0, 0), p2=(100e-9, 100e-9, 100e-9))
    mesh = df.Mesh(region=region, cell=(4e-9, 4e-9, 4e-9))
    field3 = df.Field(mesh, nvdim=3, value=m_fun, norm=Ms)
    saxs3 = mag2exp.x_ray.saxs(field3)
    m3 = abs(saxs3.array).max()

    assert np.isclose(m2 / m1, (2 / 4) ** -3, rtol=0.05)
    assert m3
    # assert np.isclose(m3 / m1, (100 / 80) ** 3)



import discretisedfield as df
import micromagneticdata as mdata
import micromagneticmodel as mm
import numpy as np
import oommfc as oc
import pytest
import xarray as xr

import mag2exp


@pytest.fixture(scope="module")
def simulation_data(tmp_path_factory):
    tmp_dir = tmp_path_factory.mktemp("sim")
    # Set up the mesh.
    mesh = df.Mesh(p1=(0, 0, 0), p2=(2e-9, 2e-9, 2e-9), n=(2, 2, 2))

    # Create the system and set up the energy, dynamics, and initial magnetisation.
    system = mm.System(name="test")
    system.energy = mm.Exchange(A=1.3e-11)
    system.dynamics = mm.Precession(gamma0=2.211e5) + mm.Damping(alpha=0.01)
    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=8e5)

    # Run the minimisation driver.
    md = oc.MinDriver()
    md.drive(system, dirname=tmp_dir)

    # Add sin wave
    system.energy += mm.Zeeman(H=(100, 0, 0), func="sin", f=10e9, t0=0)

    # Set simulation time parameters and run the time driver.
    T = 1e-9
    n = 100
    td = oc.TimeDriver()
    td.drive(system, t=T, n=n, dirname=tmp_dir)

    # Load the simulation data
    data = mdata.Data(system.name, dirname=str(tmp_dir))

    # Cleanup is automatic when the temporary directory is removed.
    return data


@pytest.fixture(scope="module")
def simulation_timedrive(simulation_data):
    return simulation_data[-1]


@pytest.fixture(scope="module")
def simulation_mindrive(simulation_data):
    return simulation_data[0]


@pytest.fixture(scope="module")
def simulation_field(simulation_data):
    return simulation_data[-1].m0


@pytest.mark.parametrize(
    "invalid_drive",
    [42, "invalid drive", None, [1, 2, 3], {}, simulation_data, simulation_field],
)
def test_invalid_drive_type(invalid_drive):
    with pytest.raises(TypeError):
        mag2exp.fmr.ringdown(invalid_drive)


def test_MinDrive(simulation_data):
    with pytest.raises(TypeError):
        mag2exp.fmr.ringdown(simulation_data[0])


@pytest.mark.parametrize(
    "invalid_field",
    [
        42,
        "invalid field",
        [1, 2, 3],
        {},
        simulation_timedrive,
        simulation_mindrive,
        simulation_data,
    ],
)
def test_invalid_init_field_type(simulation_data, invalid_field):
    with pytest.raises(TypeError):
        mag2exp.fmr.ringdown(simulation_data, init_field=invalid_field)


def test_fmr_returns_valid_arrays(simulation_timedrive):
    power, phase = mag2exp.fmr.ringdown(simulation_timedrive)

    assert isinstance(power, xr.DataArray), "power is not an xarray.DataArray."
    assert isinstance(phase, xr.DataArray), "phase is not an xarray.DataArray."

    expected_dims = {"freq_t", "x", "y", "z", "vdims"}
    assert set(power.dims) == expected_dims, "power dimensions mismatch."
    assert set(phase.dims) == expected_dims, "phase dimensions mismatch."

    expected_shape = (
        simulation_timedrive.n // 2 + 1,
        *simulation_timedrive.m0.array.shape,
    )
    assert expected_shape == power.shape, "shape of power not as expected"
    assert expected_shape == phase.shape, "shape of phase not as expected"

    assert np.all(np.isreal(power.values)), "Power array contains non-real values."
    assert np.all(np.isreal(phase.values)), "Phase array contains non-real values."

    assert phase.min() >= -np.pi, "Phase should be in the domain [-pi, pi]"
    assert phase.max() <= np.pi, "Phase should be in the domain [-pi, pi]"

    # Check largest peak is at zero frequency
    assert np.allclose(power.mean(dim=("x", "y", "z")).idxmax("freq_t"), 0)
    assert power.mean(dim=("x", "y", "z")).sel(vdims="z").max("freq_t") > 1e4

    # Check next largest peak is at excitation frequency
    assert np.allclose(
        power.mean(dim=("x", "y", "z")).isel(freq_t=slice(1, None)).idxmax("freq_t"),
        1e10,
    )


def test_fmr_with_field(simulation_timedrive, simulation_field):
    power, _ = mag2exp.fmr.ringdown(simulation_timedrive, simulation_field)

    # Check largest peak is at zero frequency
    assert np.allclose(power.mean(dim=("x", "y", "z")).idxmax("freq_t"), 0)
    assert power.mean(dim=("x", "y", "z")).sel(vdims="z").max("freq_t") < 1e4

    # Check next largest peak is at excitation frequency
    assert np.allclose(
        power.mean(dim=("x", "y", "z")).isel(freq_t=slice(1, None)).idxmax("freq_t"),
        1e10,
    )


def test_fmr_returns_valid_arrays_2d(simulation_timedrive):
    simulation_timedrive = simulation_timedrive.register_callback(
        lambda field: field.sel("z")
    )
    power, phase = mag2exp.fmr.ringdown(simulation_timedrive)

    assert isinstance(power, xr.DataArray), "power is not an xarray.DataArray."
    assert isinstance(phase, xr.DataArray), "phase is not an xarray.DataArray."

    expected_dims = {"freq_t", "x", "y", "vdims"}
    assert set(power.dims) == expected_dims, "power dimensions mismatch."
    assert set(phase.dims) == expected_dims, "phase dimensions mismatch."

    expected_shape = (
        simulation_timedrive.n // 2 + 1,
        *simulation_timedrive.m0.sel("z").array.shape,
    )
    assert expected_shape == power.shape, "shape of power not as expected"
    assert expected_shape == phase.shape, "shape of phase not as expected"

    assert np.all(np.isreal(power.values)), "Power array contains non-real values."
    assert np.all(np.isreal(phase.values)), "Phase array contains non-real values."

    assert phase.min() >= -np.pi, "Phase should be in the domain [-pi, pi]"
    assert phase.max() <= np.pi, "Phase should be in the domain [-pi, pi]"

    # Check largest peak is at zero frequency
    assert np.allclose(power.mean(dim=("x", "y")).idxmax("freq_t"), 0)
    assert power.mean(dim=("x", "y")).sel(vdims="z").max("freq_t") > 1e4

    # Check next largest peak is at excitation frequency
    assert np.allclose(
        power.mean(dim=("x", "y")).isel(freq_t=slice(1, None)).idxmax("freq_t"),
        1e10,
    )



import discretisedfield as df
import numpy as np
import pytest

import mag2exp


@pytest.mark.filterwarnings("ignore:This technique is currently under development")
def test_moke_e_field():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(1, 1, 0), norm=384e3)
    E_i = [1, 1j]
    E_f = mag2exp.moke.e_field(field, 0, 2, 1, 600e-9, E_i, mode="reflection")
    assert E_f.nvdim == 2

    E_f = mag2exp.moke.e_field(field, 0, 2, 1, 600e-9, E_i, mode="transmission")
    assert E_f.nvdim == 2

    with pytest.raises(ValueError):
        mag2exp.moke.e_field(field, 0, 2, 1, 600e-9, E_i, mode="blah")


@pytest.mark.filterwarnings("ignore:This technique is currently under development")
def test_moke_intensity():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=384e3)
    E_i = [1, 1j]
    intensity = mag2exp.moke.intensity(field, 0, 1, 0, 600e-9, E_i, mode="reflection")

    assert np.allclose(intensity.array, 0)

    field = df.Field(mesh, nvdim=3, value=(1, 1, 0), norm=384e3)
    E_i = [1, 0]
    intensity = mag2exp.moke.intensity(field, 0, 1, 0, 600e-9, E_i, mode="reflection")

    assert np.allclose(intensity.array, 0)

    mag2exp.moke.intensity(field, 0, 1, 0, 600e-9, E_i, mode="transmission")


@pytest.mark.filterwarnings("ignore:This technique is currently under development")
def test_moke_angle():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=384e3)
    angle = mag2exp.moke.kerr_angle(field, 0, 2, 1, 600e-9)

    assert angle.array.all() != 0


@pytest.mark.filterwarnings("ignore:This technique is currently under development")
def test_moke_fwhm():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    field = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=384e3)
    E_i = [1, 1j]
    mag2exp.moke.intensity(
        field, 0, 1, 0, 600e-9, E_i, mode="reflection", fwhm=(1e-9, 1e-9)
    )



"""MFM submodule.

Module for calculation of Magnetic Force Microscopy related quantities.
"""

import micromagneticmodel as mm

import mag2exp


def phase_shift(field, /, tip_m=(0, 0, 0), quality=650, k=3, tip_q=0, fwhm=None):
    r"""Calculation of the phase shift of an MFM tip.

    The contrast in MFM images originates from the magnetic interaction between
    the magnetic tip of an oscillating cantilever and the samples stray field.
    As MFM is based on the stray field outside of a sample, an 'airbox' method
    should be used. This mean that the saturation magnetisation should be set
    to zero in the region outside the sample in which we wish to perform these
    MFM measurements.

    The magnetic cantilever is driven to oscillate near its resonant frequency
    when there is no stray field. In the presence of a stray magnetic field the
    phase shift of MFM cantilever is given by

    .. math::
        \Delta \phi = \frac{Q\mu_0}{k} \left( q \frac{\partial
        {\bf H}_{sz}}{\partial z} + {\bf M}_t \cdot
        \frac{\partial^2{\bf H}_{s}}{\partial z^2} \right),

    where
    :math:`Q` is the quality factor of the cantilever,
    :math:`k` is the spring constant of the cantilever in
    :math:`\textrm{Nm}^{-1}`, :math:`q` is the effective magnetic monopole
    moment of the tip in :math:`\textrm{Am}^{-2}`,
    :math:`{\bf M}_t` is the effective magnetic dipole moment of the tip in
    :math:`\textrm{Am}^{-1}` and,
    :math:`{\bf H}_{sz}` is the magnetic field due to the sample in
    :math:`\textrm{Am}^{-1}`.


    Parameters
    ----------
    field : discretisedfield.Field
        Magnetisation field.

    tip_m : numbers.Real, array_like
        The effective magnetic dipole moment of the tip in the :math:`x`,
        :math:`y`, and :math:`z` directions, respectively. Values are given in
        units of :math:`\textrm{Am}^{-1}`.

    quality : numbers.Real
        The quality factor of the tip. Should be a real, positive number.

    k : numbers.Real
        The spring factor of the MFM cantilever. This should be a real,
        positive, non-zero number.

    tip_q : numbers.Real
        The effective magnetic monopole moment of the tip in units of
        :math:`\textrm{Am}^{-2}`.

    Returns
    -------
    discretisedfield.Field
        Phase shift of MFM cantilever.

    Examples
    --------

    .. plot::
        :context: close-figs

        1. Visualising MFM with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
        ...                p2=(25e-9, 25e-9, 50e-9),
        ...                cell=(1e-9, 1e-9, 2e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     if x < -2e-9:
        ...         return (0, 0, 1)
        ...     elif x < 2e-9:
        ...         return (0, 1, 0)
        ...     else:
        ...         return (0, 0, -1)
        >>> def Ms_fun(pos):
        ...     x, y, z = pos
        ...     if (z < 0):
        ...         return 384e3
        ...     else:
        ...         return 0
        >>> field = df.Field(mesh, nvdim=3, value=v_fun, norm=Ms_fun)
        >>> ps = mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 1e-16))
        >>> ps.sel(z=10e-9).mpl.scalar()
        >>> ps.sel(z=40e-9).mpl.scalar()


    .. plot::
        :context: close-figs

        2. Visualising in-plane MFM with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
        ...                p2=(25e-9, 25e-9, 50e-9),
        ...                cell=(1e-9, 1e-9, 2e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     if x < -2e-9:
        ...         return (0, 0, 1)
        ...     elif x < 2e-9:
        ...         return (0, 1, 0)
        ...     else:
        ...         return (0, 0, -1)
        >>> def Ms_fun(pos):
        ...     x, y, z = pos
        ...     if (z < 0):
        ...         return 384e3
        ...     else:
        ...         return 0
        >>> system = mm.System(name='Box2')
        >>> system.energy = mm.Demag()
        >>> system.m = df.Field(mesh, nvdim=3, value=v_fun, norm=Ms_fun)
        >>> ps = mag2exp.mfm.phase_shift(system.m, tip_m=(1e-16, 0, 0))
        >>> ps.sel(z=10e-9).mpl.scalar()
        >>> ps.sel(z=40e-9).mpl.scalar()


    .. plot::
        :context: close-figs

        3. Visualising monopole moment MFM with ``matplotlib``.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(-25e-9, -25e-9, -2e-9),
        ...                p2=(25e-9, 25e-9, 50e-9),
        ...                cell=(1e-9, 1e-9, 2e-9))
        >>> def v_fun(point):
        ...     x, y, z = point
        ...     if x < -2e-9:
        ...         return (0, 0, 1)
        ...     elif x < 2e-9:
        ...         return (0, 1, 0)
        ...     else:
        ...         return (0, 0, -1)
        >>> def Ms_fun(pos):
        ...     x, y, z = pos
        ...     if (z < 0):
        ...         return 384e3
        ...     else:
        ...         return 0
        >>> system = mm.System(name='Box2')
        >>> system.energy = mm.Demag()
        >>> system.m = df.Field(mesh, nvdim=3, value=v_fun, norm=Ms_fun)
        >>> ps = mag2exp.mfm.phase_shift(system.m, tip_q=1e-9)
        >>> ps.sel(z=10e-9).mpl.scalar()
        >>> ps.sel(z=40e-9).mpl.scalar()
    """

    if k <= 0:
        msg = "`k` has to be a positive non-zero number."
        raise RuntimeError(msg)

    stray_field = mag2exp.util.calculate_demag_field(field)
    dh_dz = stray_field.diff("z", order=1)
    d2h_dz2 = stray_field.diff("z", order=2)
    phase_shift = (quality * mm.consts.mu0 / k) * (tip_q * dh_dz.z + d2h_dz2 @ tip_m)
    if fwhm is not None:
        phase_shift = mag2exp.util.gaussian_filter(phase_shift, fwhm=fwhm)
    return phase_shift



import discretisedfield as df
import micromagneticmodel as mm
import numpy as np
import pytest

import mag2exp


def test_relativistic_wavelength():
    with pytest.warns(RuntimeWarning, match="divide by zero"):
        assert mag2exp.ltem.relativistic_wavelength(0) == float("inf")

    assert np.isclose(mag2exp.ltem.relativistic_wavelength(50000), 5.3408e-12)


def test_ltem_phase():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    phase, ft_phase = mag2exp.ltem.phase(field)
    assert (phase.real.array == 0).all()
    assert (phase.imag.array == 0).all()


def test_ltem_phase_neel():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (1, 0, 0)
        else:
            return (-1, 0, 0)

    field = df.Field(mesh, nvdim=3, value=f_val)
    phase, ft_phase = mag2exp.ltem.phase(field)
    assert (phase.real.array == 0).all()
    assert (phase.imag.array == 0).all()


def test_ltem_phase_bloch():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 1, 0)
        else:
            return (0, -1, 0)

    field = df.Field(mesh, nvdim=3, value=f_val)
    phase, ft_phase = mag2exp.ltem.phase(field)
    assert (phase.real.array != 0).any()
    assert phase.imag.allclose(df.Field(mesh=phase.mesh, nvdim=1))


def test_defocus_image():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    phase, ft_phase = mag2exp.ltem.phase(field)
    with pytest.raises(RuntimeError):
        mag2exp.ltem.defocus_image(phase)


def test_defocus_image_zero_df():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 1, 0)
        else:
            return (0, -1, 0)

    field = df.Field(mesh, nvdim=3, value=f_val)
    phase, ft_phase = mag2exp.ltem.phase(field)
    dfi = mag2exp.ltem.defocus_image(phase, cs=0, df_length=0, voltage=300e3)
    assert (dfi.array == 1).all()


def test_defocus_image_df():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 1, 0)
        else:
            return (0, -1, 0)

    field = df.Field(mesh, nvdim=3, value=f_val)
    phase, ft_phase = mag2exp.ltem.phase(field)
    dfi = mag2exp.ltem.defocus_image(phase, cs=0, df_length=0.2e-3, voltage=300e3)
    assert np.allclose(dfi.array, 1)


def test_integrated_magnetic_flux_density():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 1, 0)
        else:
            return (0, -1, 0)

    field = df.Field(mesh, nvdim=3, value=f_val)
    phase, _ = mag2exp.ltem.phase(field)
    imf = mag2exp.ltem.integrated_magnetic_flux_density(phase)
    assert (imf.array != 0).any()


def test_phase_analytical():
    Ms = 1.1e6
    region = df.Region(p1=(0, 0, 0), p2=(150e-9, 100e-9, 10e-9))
    mesh = df.Mesh(region=region, cell=(1e-9, 1e-9, 1e-9))

    def m_fun(pos):
        x, y, z = pos
        q = 30e-9
        return [0, np.sin(2 * np.pi * x / q), np.cos(2 * np.pi * x / q)]

    field = df.Field(mesh, nvdim=3, value=m_fun, norm=Ms)
    phase, _ = mag2exp.ltem.phase(field)

    def analytical_sol(pos):
        x, y = pos
        q = 30e-9
        analytical = -((mm.consts.e * mm.consts.mu0 * Ms * 10e-9) / (mm.consts.h))
        analytical *= np.cos(2 * np.pi * x / q) * q
        analytical *= (np.sin(2 * np.pi * (0.5e-9) / q)) / (2 * np.pi * 0.5e-9 / q)
        return analytical

    an_phase = df.Field(phase.mesh, nvdim=1, value=analytical_sol)
    cut = phase.line(p1=(30e-9, 50e-9), p2=(120e-9, 50e-9), n=90)
    an_cut = an_phase.line(p1=(30e-9, 50e-9), p2=(120e-9, 50e-9), n=90)
    assert np.isclose(
        cut.data["v"].to_numpy(), an_cut.data["v"].to_numpy(), rtol=1e-3
    ).all()



"""Simulate experimental measurements."""

import importlib

import pytest

from . import fmr as fmr
from . import ltem as ltem
from . import magnetometry as magnetometry
from . import mfm as mfm
from . import moke as moke
from . import quick_plots as quick_plots
from . import sans as sans
from . import util as util
from . import x_ray as x_ray

__version__ = importlib.metadata.version(__package__)


def test():
    """Run all package tests.

    Examples
    --------
    1. Run all tests.

    >>> import mag2exp
    ...
    >>> # mag2exp.test()

    """
    return pytest.main(["-v", "--pyargs", "mag2exp", "-l"])  # pragma: no cover



import discretisedfield as df
import micromagneticmodel as mm
import numpy as np
import oommfc as oc
import scipy.ndimage


def gaussian_filter(field, /, fwhm):
    """Gaussian filter.

    Convolution of a field with a 1 dimensional filter.

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.
    fwhm : array_like, optional
        If specified, convolutes the output image with a 2 Dimensional Gaussian
        with the full width half maximum (fwhm) specified.

    Returns
    -------
    discretisedfield.Field
        Convoluted field.

    Raises
    ------
    RuntimeError
        Gaussian filter only supports fields with field.nvdim=1.
    """
    if field.nvdim > 1:
        msg = "Gaussian filter only supports fields with field.nvdim=1."
        raise RuntimeError(msg)
    sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
    if field.mesh.region.ndim == 2:
        sigma = (
            sigma[0] / field.mesh.cell[0],
            sigma[1] / field.mesh.cell[1],
        )
        value = scipy.ndimage.gaussian_filter(field.array.squeeze(), sigma=sigma)[
            ..., np.newaxis
        ]
    else:
        sigma = [sigma[i] / field.mesh.cell[i] for i in range(3)]
        value = scipy.ndimage.gaussian_filter(field.array.squeeze(), sigma=sigma)[
            ..., np.newaxis
        ]

    return df.Field(field.mesh, nvdim=1, value=value)


def calculate_demag_field(field):
    """Calculate demagnetisation field.

    Calculate demagnetisation field using OOMMF.

    Parameters
    ----------
    field : discretisedfield.field
        Magnetisation field.

    Returns
    -------
    discretisedfield.Field
        Demagnetisation field.

    """
    system = mm.System(name="demag_calculation")
    system.energy = mm.Demag()
    system.m = field
    return oc.compute(system.energy.demag.effective_field, system, verbose=0)



"""FMR submodule.

Module for calculation of ferromagnetic resonance related
quantities.
"""

from typing import Optional

import discretisedfield as df
import micromagneticdata as mdata
import numpy as np
import scipy.fft as fft
import xarray as xr


def ringdown(
    drive: mdata.Drive, init_field: Optional[df.Field] = None
) -> tuple[xr.DataArray, xr.DataArray]:
    r"""
    Compute the Ferromagnetic Resonance (FMR) power and phase spectra
    using the ringdown method.

    This function extracts the magnetisation from the
    provided drive object, optionally subtracts an initial field,
    and applies a Fourier transform to obtain the power
    and phase spectra.

    The Discrete Fourier Transform is computed at each spatial grid point for the
    :math:`k`-th component of the magnetisation.
    The corresponding power spectrum is given by:

    .. math::

        S_k(r_j, f) = \frac{1}{N} \left| F_k(r_j, f) \right|^2

    where :math:`F_k(r_j, f)` is the Fourier transform of the magnetisation component
    :math:`M_k(r_j, t)`.

    The phase spectrum, which provides spatially resolved phase information,
    is given by:

    .. math::

        \phi_k(r_j, f) = \arg F_k(r_j, f).

    where :math:`\arg` is the phase angle of the complex Fourier coefficient.

    For further details on numerical FMR analysis in micromagnetics see,
    A. Baker et al., J. Magn. Magn. Mater. 421, 428 (2017)

    Parameters
    ----------
    drive : micromagneticdata.Drive
        A micromagnetic drive object containing time-dependent
        magnetisation data.
    init_field : discretisedfield.Field, optional
        A reference field whose orientation is subtracted before processing.

    Returns
    -------
    power : xr.DataArray
        The computed power spectrum (squared magnitude of the FFT).
    phase : xr.DataArray
        The computed phase spectrum (FFT phase angle).

    Example
    -------
    .. plot::
        :context: close-figs

        1. Calculating the FMR power and phase using
        the ringdown method.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import oommfc as oc
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(0, 0, 0),
        ...                p2=(2e-9, 2e-9, 2e-9),
        ...                n=(2, 2, 2))
        >>> system = mm.System(name="test")
        >>> system.energy = (mm.Exchange(A=1.3e-11)
        ...    + mm.Zeeman(H=(100, 0, 0), func="sin", f=10e9, t0=0))
        >>> system.dynamics = (mm.Precession(gamma0=2.211e5)
        ...                    + mm.Damping(alpha=0.01))
        >>> system.m = df.Field(mesh, nvdim=3,
        ...                     value=(0, 0, 1), norm=8e5)
        >>> td = oc.TimeDriver()
        >>> td.drive(system, t=1e-9, n=100)
        Running OOMMF ...
        >>> drive = mdata.Data(system.name)[-1]
        >>> power, phase = mag2exp.fmr.ringdown(drive)

    .. plot::
        :context: close-figs

        2. Use the orientation to calculate the FMR power
        and phase using the ringdown method.

        >>> import discretisedfield as df
        >>> import micromagneticmodel as mm
        >>> import oommfc as oc
        >>> import mag2exp
        >>> mesh = df.Mesh(p1=(0, 0, 0),
        ...                p2=(2e-9, 2e-9, 2e-9),
        ...                n=(2, 2, 2))
        >>> system = mm.System(name="test")
        >>> system.energy = (mm.Exchange(A=1.3e-11)
        ...    + mm.Zeeman(H=(100, 0, 0), func="sin", f=10e9, t0=0))
        >>> system.dynamics = (mm.Precession(gamma0=2.211e5)
        ...                    + mm.Damping(alpha=0.01))
        >>> system.m = df.Field(mesh, nvdim=3,
        ...                     value=(0, 0, 1), norm=8e5)
        >>> td = oc.TimeDriver()
        >>> td.drive(system, t=1e-9, n=100)
        Running OOMMF ...
        >>> drive = mdata.Data(system.name)[-1]
        >>> drive = drive.register_callback(
        ...    lambda field: field.orientation)
        >>> power, phase = mag2exp.fmr.ringdown(drive)

    """
    if not isinstance(drive, mdata.Drive):
        raise TypeError(
            "The 'drive' parameter must be an instance of "
            f"micromagneticdata.Drive, not {type(drive)=}"
        )
    if init_field is not None and not isinstance(init_field, df.Field):
        raise TypeError(
            "The 'init_field' parameter must be an instance of "
            f"micromagneticdata.Field if provided, not {type(init_field)=}."
        )

    if drive.x != "t":
        raise TypeError("The drive data must have a 't' (time) coordinate.")

    # Validate time step uniformity
    t_values = drive.table.data["t"]
    if t_values.size < 2:
        raise ValueError("Insufficient time points to compute a FFT.")
    dt_array = np.diff(t_values)
    if not np.allclose(dt_array, dt_array[0], atol=0.01 * dt_array[0]):
        raise ValueError("Time steps in the drive data are not uniform.")
    dt = dt_array[0]

    data_xarr = drive.to_xarray()

    if init_field is not None:
        data_xarr -= init_field.to_xarray()

    # Compute FFT frequencies and FFT along the time axis
    # (The first dimension of drive.to_xarray()).
    num_time_points = data_xarr.shape[0]
    freq = fft.rfftfreq(num_time_points, d=dt)
    fft_values = fft.rfft(data_xarr.data, axis=0, norm="ortho")

    # Keep all coordinates except the time coordinate ('t') to the FFT output.
    # Done like this as the names of dims can vary.
    fft_coords = {"freq_t": freq}
    fft_dims = ["freq_t"]
    for dim in data_xarr.dims:
        if dim != "t":
            fft_coords[dim] = data_xarr.coords[dim]
            fft_dims.append(dim)

    fft_xarr = xr.DataArray(fft_values, coords=fft_coords, dims=fft_dims, name="fft")

    power = np.abs(fft_xarr) ** 2
    power.name = "power"
    phase = np.arctan2(fft_xarr.imag, fft_xarr.real)
    phase.name = "phase"

    return power, phase



# import pytest
import discretisedfield as df
import micromagneticmodel as mm
import numpy as np

import mag2exp


def test_magnetisation_analytical():
    mesh = df.Mesh(p1=(0, 0, 0), p2=(6e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9))
    field = df.Field(mesh, nvdim=3, value=(1, 1, 1))
    mag = mag2exp.magnetometry.magnetisation(field)
    assert np.isclose(mag, 1).all()


def test_torque_analytical_demag():
    mesh = df.Mesh(p1=(0, 0, 0), p2=(6e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9))
    field = df.Field(mesh, nvdim=3, value=(0, 1e5, 0))
    H = (0, 0, 1e5)
    torque = mag2exp.magnetometry.torque(field, H)
    assert np.isclose(torque[0], mm.consts.mu0 * 1e10)
    assert np.isclose(torque[1], 0)
    assert np.isclose(torque[2], 0)


def test_torque_analytical__parallel_demag():
    mesh = df.Mesh(p1=(0, 0, 0), p2=(6e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9))
    field = df.Field(mesh, nvdim=3, value=(0, 1e5, 0))
    H = (0, 1e5, 0)
    torque = mag2exp.magnetometry.torque(field, H)
    assert np.isclose(torque, 0).all()



import matplotlib
import matplotlib.pyplot as plt
import ubermagutil.units as uu

from . import ltem, mfm, sans, x_ray


def ltem_phase(field, /, kcx=0.1, kcy=0.1):
    r"""Quickplot of the magnetic phase shift.

    The phase is calculated using the :code:`ltem.phase`
    function and plotted using :code:`mpl.scalar`.

    .. seealso:: :py:func:`mag2exp.ltem.phase`
    """
    phase, _ = ltem.phase(field, kcx=kcx, kcy=kcy)
    phase.mpl.scalar(
        cmap="gray", interpolation="spline16", colorbar_label=r"$\phi$ (radians)"
    )


def ltem_ft_phase(field, /, kcx=0.1, kcy=0.1):
    r"""Quickplot of the magnetic phase shift in Fourier space.

    The Fourier transform of the phase is calculated using the
    :code:`ltem.phase` function and plotted using
    :code:`mpl.scalar`.

    .. seealso:: :py:func:`mag2exp.ltem.phase`
    """
    _, ft_phase = ltem.phase(field, kcx=kcx, kcy=kcy)
    fig, ax = plt.subplots()
    (ft_phase.conjugate * ft_phase).real.mpl.scalar(
        ax=ax,
        cmap="gray",
        interpolation="spline16",
        colorbar_label=(
            r"$\left\vert\widetilde{\phi}\right\vert^2$" "(radians$^{-2}$)"
        ),
    )
    multiplier = uu.si_max_multiplier(ft_phase.mesh.region.edges)
    ax.add_patch(
        matplotlib.patches.Ellipse(
            xy=(0, 0),
            width=ft_phase.mesh.cell[0] * kcx * 2 / multiplier,
            height=ft_phase.mesh.cell[1] * kcy * 2 / multiplier,
            edgecolor="red",
            facecolor="none",
            linewidth=3,
            label="Tikhonov filter",
        )
    )
    ax.legend(frameon=True)


def ltem_defocus(
    field, /, kcx=0.1, kcy=0.1, cs=0, df_length=0.2e-3, voltage=None, wavelength=None
):
    r"""Quickplot of the LTEM defocus image.

    The phase is calculated using the :code:`ltem.phase`
    function and propagated to the image plane using
    :code:`ltem.defocus_image`.	This is then plotted using
    :code:`mpl.scalar`.

    .. seealso::
        :py:func:`mag2exp.ltem.phase`
        :py:func:`mag2exp.ltem.defocus_image`
    """
    phase, _ = ltem.phase(field, kcx=kcx, kcy=kcy)
    defocus = ltem.defocus_image(
        phase, cs=cs, df_length=df_length, voltage=voltage, wavelength=wavelength
    )
    defocus.mpl.scalar(
        cmap="gray", interpolation="spline16", colorbar_label="Normalised Intensity"
    )


def ltem_integrated_mfd(field, /, kcx=0.1, kcy=0.1):
    r"""Quickplot of the LTEM integrated magnetic flux density.

    The phase is calculated using the :code:`ltem.phase`
    function and the integrated magnetic flux density
    :code:`ltem.integrated_magnetic_flux_density`.
    This is then plotted using :code:`mpl.lightness` and :code:`mpl.vector`.

    .. seealso::
        :py:func:`mag2exp.ltem.phase`
        :py:func:`mag2exp.ltem.integrated_magnetic_flux_density`

    """
    phase, _ = ltem.phase(field, kcx=kcx, kcy=kcy)
    imf = ltem.integrated_magnetic_flux_density(phase)
    fig, ax = plt.subplots()
    imf.mpl.lightness(
        ax=ax,
        clim=[0, 0.5],
        interpolation="spline16",
        colorwheel_args=dict(width=0.75, height=0.75),
        colorwheel_xlabel=r"$m_x$",
        colorwheel_ylabel=r"$m_y$",
    )
    imf.mpl.vector(ax=ax, vdims=["x", "y"], use_color=False, color="w")


def mfm_phase_shift(
    field, /, tip_m=(0, 0, 0), quality=650, k=3, tip_q=0, fwhm=None, z0=0
):
    r"""Quickplot of the magnetic phase shift.

    The phase is calculated using the :code:`ltem.phase`
    function and plotted using :code:`mpl.scalar`.

    .. seealso:: :py:func:`mag2exp.mfm.phase_shift`
    """
    phase_shift = mfm.phase_shift(
        field, tip_m=tip_m, quality=quality, k=k, tip_q=tip_q, fwhm=fwhm
    )
    phase_shift_p = phase_shift.sel(z=z0)
    phase_shift_p.mpl.scalar(
        interpolation="spline16", colorbar_label=r"Phase shift (radians.)"
    )


def x_ray_holography(field, /, fwhm=None):
    r"""Quickplot of the magnetic phase shift.

    The phase is calculated using the :code:`ltem.phase`
    function and plotted using :code:`mpl.scalar`.

    .. seealso:: :py:func:`mag2exp.x_ray.holography`
    """
    holo = x_ray.holography(field, fwhm=fwhm)
    holo.mpl.scalar(
        cmap="RdBu",
        symmetric_clim=True,
        interpolation="spline16",
        colorbar_label=r"Integrated Magnetisation (A)",
    )


def saxs(field):
    r"""Quickplot of the small angle x-ray scattering pattern.

    The small angle x-ray scattering pattern is calculated using the
    :code:`x_ray.saxs` function and plotted using :code:`mpl.scalar`.

    .. seealso:: :py:func:`mag2exp.x_ray.saxs`

    """
    cs = x_ray.saxs(field)
    cs.mpl.scalar(
        cmap="gray", interpolation="spline16", colorbar_label=r"Intensity (arb.)"
    )


def sans_cross_section(field, /, method, polarisation=(0, 0, 1)):
    r"""Quickplot of the small angle neutron scattering pattern.

    The small angle neutron scattering pattern is calculated using the
    :code:`sans.cross_section` function and plotted using :code:`mpl.scalar`.

    .. seealso:: :py:func:`mag2exp.sans.cross_section`

    """
    cs = sans.cross_section(field, method=method, polarisation=polarisation)
    cs.sel(k_z=0).mpl.scalar(
        cmap="gray", interpolation="spline16", colorbar_label=r"Intensity (arb.)"
    )


def sans_chiral_function(field, /, polarisation=(0, 0, 1)):
    r"""Quickplot of the small angle neutron scattering chiral function.

    The small angle neutron scattering chiral function is calculated using the
    :code:`sans.chiral_function` function and the imaginary component
    plotted using :code:`mpl.scalar`.

    .. seealso:: :py:func:`mag2exp.sans.chiral_function`
    """
    cf = sans.chiral_function(field, polarisation=polarisation)
    cf.sel(k_z=0).mpl.scalar(
        cmap="gray", interpolation="spline16", colorbar_label=r"Cross section (arb.)"
    )



# import pytest
import discretisedfield as df

import mag2exp


def test_quick_plots_ltem_phase():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.quick_plots.ltem_phase(field)


def test_quick_plots_ltem_ft_phase():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.quick_plots.ltem_ft_phase(field)


def test_quick_plots_ltem_defocus():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.quick_plots.ltem_defocus(field, voltage=300e3)


def test_quick_plots_ltem_integrated_mfd():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )

    def v_fun(point):
        x, y, z = point
        if x < -2e-9:
            return (0, 0, 1)
        elif x < 2e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    field = df.Field(mesh, nvdim=3, value=v_fun)
    mag2exp.quick_plots.ltem_integrated_mfd(field)


def test_quick_plots_mfm_phase_shift():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )

    def v_fun(point):
        x, y, z = point
        if x < -2e-9:
            return (0, 0, 1)
        elif x < 2e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    def Ms_fun(pos):
        x, y, z = pos
        if z < 0:
            return 384e3
        else:
            return 0

    field = df.Field(mesh, nvdim=3, value=v_fun, norm=Ms_fun)
    mag2exp.quick_plots.mfm_phase_shift(field)


def test_quick_plots_x_ray_holography():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.quick_plots.x_ray_holography(field)


def test_quick_plots_saxs():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.quick_plots.saxs(field)


def test_quick_plots_sans_cross_section():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.quick_plots.sans_cross_section(field, method="unpol")


def test_quick_plots_sans_chiral_function():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -1e-9), p2=(5e-9, 4e-9, 1e-9), cell=(2e-9, 1e-9, 0.5e-9)
    )
    field = df.Field(mesh, nvdim=3, value=(0, 0, 1))
    mag2exp.quick_plots.sans_chiral_function(field)



import discretisedfield as df
import pytest

import mag2exp


def test_mfm_phase_tip_m():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 0, 1)
        elif x < 1e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    def Ms_fun(pos):
        x, y, z = pos
        if z < 0:
            return 384e3
        else:
            return 0

    field = df.Field(mesh, nvdim=3, value=f_val, norm=Ms_fun)
    ps = mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 1e-16))
    assert (ps.array != 0).any()
    ps = mag2exp.mfm.phase_shift(field, tip_m=(0, 1e-16, 0))
    assert (ps.array != 0).any()
    ps = mag2exp.mfm.phase_shift(field, tip_m=(1e-16, 0, 0))
    assert (ps.array != 0).any()


def test_mfm_phase_tip_q():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 0, 1)
        elif x < 1e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    def Ms_fun(pos):
        x, y, z = pos
        if z < 0:
            return 384e3
        else:
            return 0

    field = df.Field(mesh, nvdim=3, value=f_val, norm=Ms_fun)
    ps = mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 0), quality=650, k=3, tip_q=1e-6)
    assert (ps.array != 0).any()


def test_mfm_phase_no_tip():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 0, 1)
        elif x < 1e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    def Ms_fun(pos):
        x, y, z = pos
        if z < 0:
            return 384e3
        else:
            return 0

    field = df.Field(mesh, nvdim=3, value=f_val, norm=Ms_fun)
    ps = mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 0), quality=650, k=3, tip_q=0)
    assert (ps.array == 0).all()


def test_mfm_phase_quality():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 0, 1)
        elif x < 1e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    def Ms_fun(pos):
        x, y, z = pos
        if z < 0:
            return 384e3
        else:
            return 0

    field = df.Field(mesh, nvdim=3, value=f_val, norm=Ms_fun)
    ps = mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 1e-16), quality=0)
    assert (ps.array == 0).all()


def test_mfm_phase_k():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 0, 1)
        elif x < 1e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    def Ms_fun(pos):
        x, y, z = pos
        if z < 0:
            return 384e3
        else:
            return 0

    field = df.Field(mesh, nvdim=3, value=f_val, norm=Ms_fun)
    with pytest.raises(RuntimeError):
        mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 1e-16), k=0)
    with pytest.raises(RuntimeError):
        mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 1e-16), k=-3)


def test_mfm_phase_fwhm():
    mesh = df.Mesh(
        p1=(-5e-9, -4e-9, -2e-9), p2=(5e-9, 4e-9, 6e-9), cell=(2e-9, 1e-9, 2e-9)
    )

    def f_val(position):
        x, y, z = position
        if x < 0:
            return (0, 0, 1)
        elif x < 1e-9:
            return (0, 1, 0)
        else:
            return (0, 0, -1)

    def Ms_fun(pos):
        x, y, z = pos
        if z < 0:
            return 384e3
        else:
            return 0

    field = df.Field(mesh, nvdim=3, value=f_val, norm=Ms_fun)
    mag2exp.mfm.phase_shift(field, tip_m=(0, 0, 1e-16), fwhm=(1e-9, 1e-9, 1e-9))
